/* Solver for the game of Patience as seen in EXAPUNKS by Zachtronics.        */
/* (c) 2018-19 Ian Collier                                                    */

/* Input and output are in the form of stringified Python arrays because this
   will be interfaced with Python scripts.
   Flags: -v    print verbose progress messages on stderr
          -vvN  (where N is a number) print progress messages up to depth N
          -n    do not use the cache
   Input: list of list of integers.  Each integer can be in decimal or
          "0x" followed by hex digits, and represents a card (value 1 to 4
          for the four suits - we don't care in what order - or 6 to 10
          for the numbers plus 0x80 (aka REDCARD) if the number is red).
          The outer list is a list of the nine columns in the game; each
          column is a list of cards in that column (top first).  A tenth
          column containing a single card may be supplied, and this represents
          the free cell.
   Output: either "None" (if no solution found) or a list of list of Tuples.
          The outer list is a list of moves; each move is a list [(w,x), (y,z)]
          where (w,x) is the column and row of the card to be moved and (y,z) is
          the column and row of the destination to move to (zero-based
          coordinates from left-to-right and top-to-bottom).

Strategy:

This will be a simple backtrack search.  For every possible legal move from
the current position: make the move, call the search recursively to find out
whether there is a solution from there, and undo the move.  If the answer was
no, go on to consider the next move, and if there is no next move, return "no".
If the answer was yes, we can either return "yes" or carry on to see if there
is a shorter solution.

Each time the answer is "yes", the move that leads to the solution will be
recorded in an array.  If the answer is "no", the array won't be touched.
So once the top-level solve function returns "yes" this represents a "yes"
from every call below it, and every move in the solution will have been
recorded.

The solution space for this game is BIG, so we use three strategies to
cut it down:
 1. Try to find a solution without using the free cell, because if it's
    possible then this will be much quicker as it cuts down the number of
    legal moves at each stage.  If that fails then we must try again but
    with the free cell in use.
 2. Extra restrictions on what moves are legal, without changing the accessible
    problem space.  We will insist that if a sequence exists in a column of
    cards, the whole sequence must be moved at once.  There is no point in
    taking a black 6 off a red 7 and putting it on a different red 7, and
    almost certainly no point in putting the 6 in an empty column or in the
    free cell.  If two different moves are legal at the same time and could
    be carried out in either order, we will insist that the left-hand move
    is carried out first.  When using the free cell, we will not move a card
    to the cell and then put it straight back on the board; nor vice versa.
 3. Every board position we reach will be cached.  The object of the cache
    is to answer the question "Can this position be solved in at most N moves?"
    If we encounter this same board position again with at most N moves left,
    if the cache says the answer was "no" then we can return "no" without
    re-computing it.  (If the answer was "yes" then we will re-compute it,
    because the cache does not save the winning moves.)

In addition, the maximum number of moves allowed will be limited, to prevent
infinite cycles and wild goose chases.  The initial number for this should be
large enough that we think most or all games can be solved in that many moves.

There are two ways to count the length of a solution: (a) the number of times
a card or cards are moved (which is also the length of the output array), and
(b) the number of individual cards moved.  So if you move an 8-7-6 sequence
from one column to another, this would be 1 in scheme (a) but 3 in scheme (b).
For computing the "shortest" solution, this solver uses count method (b).  This
makes it favour building up a sequence in one place rather than moving the
whole sequence to collect more cards.  For example, if there is a 9, an 8-7
and a 6, we put the 8-7 on to the 9 before moving the 6, rather than moving
the 6 and then moving the 8-7-6 sequence.

*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<memory.h>
#include<time.h>

#define REDCARD 128            /* a bit in the input representing a red number */
#define CELL 9                 /* number of columns; also the coord of the free cell */
#define MAXROWS 8              /* maximum no. of cards that can be in each column */

#define MAXDEPTH 90            /* maximum recursion depth we want to go to */
#define MAXMOVES 90            /* maximum number of moves we want to look for */

#ifndef HASHBITS
# define HASHBITS 22           /* Default size of the hash: 2^22 entries */
#endif
#define HASHSIZE (1<<HASHBITS)
#define HASHMASK (HASHSIZE-1)
#define HASHFULL (HASHSIZE/2)  /* Consider clearing the hash when it has this many entries */

/* The MODULUS and GENERATOR are used in the hash function to generate a
   pseudo-random number based on each board position.  The MODULUS is a
   prime number with enough bits to fill the hash table, and the GENERATOR
   is a primitive root with respect to that modulus.  Conveniently, 2^31-1
   is a prime number, and the generator here is that used in the MINSTD
   Lehmer RNG and is equal to 7^5. */
#define MODULUS 2147483647
#define GENERATOR 16807

static int OPTTIME=8;          /* no. of seconds to spend looking for a better solution */
static int MAXTIME=600;        /* no. of seconds to spend looking for any solution */

/* The cache depth variables allow us to specify the recursion depths at
   which the cache will be used.  We might want to limit the maximum depth
   because at that depth it's more efficient to just calculate the solution
   rather than waste memory in the cache.  But experience shows that on a 
   modern computer system it's usually possible to just cache everything
   (unless running with "-n" in which case cmindepth will be set so that
   no caching occurs).  The cache is not supposed to alter the result of
   the computation, so using "-n" or changing the cache depths allow us
   to verify by experiment that this is the case. */
static int cmindepth=0;
static int cachedepth=80;

/* The cache answers the question "Can this position be solved in at most
   N moves?" and will be stored in a hash table.  Each item in the table
   will be a linked list of buckets, each one looking like this: */
typedef struct _hashitem {
    struct _hashitem *next;   /* pointer to next bucket, or NULL */
    unsigned long hash;       /* full value of the hash function */
    int max;                  /* max number of moves we asked about */
    int result;               /* the cached answer */
    unsigned char b[1];       /* the first byte of a cached copy of the board */
} Hash;

static Hash **hash;           /* the root of the hash table */
static unsigned long hashsize;/* number of hash entries in the table */

/* The board will be a 2-dimensional array of cards.  The [0] element of each
   column will be the count N of cards in that column, and then elements [1] to
   [N] will be the cards themselves.  The free cell is the last column and is
   only allowed to hold one card or no cards. */
unsigned char board[CELL+1][MAXROWS+1];  

static int verbose=0;         /* whether to print verbose messages */
static int vdepth=4;          /* maximum recusion depth to print the msgs */

/* The myalloc function simply calls malloc to allocate memory, but saves
 * the bother of checking the return value by aborting if malloc failed.
 */
void *myalloc(size_t size) {
    void *m=malloc(size);
    if (m) return m;
    fprintf(stderr, "failed to allocate block of size %lu\n", (unsigned long)size);
    exit(1);
}

/* The powmod function calculates (y^x) mod(MODULUS) and is used in the
 * hash function.  We are assuming that "long" is long enough to hold the
 * modulus and "long long" is long enough to hold two longs multiplied
 * together.
 *
 * Square-and-multiply method.  Check each bit of x in turn, square the
 * answer so far and multiply by y if the bit was a '1'.
 */
unsigned long powmod(unsigned long y, unsigned long x) {
    unsigned long answer=1;
    int i;
    for (i=0; i<32; i++) {
        if (i>0 && answer>1)
            answer = ((long long)answer*(long long) answer)%MODULUS;
        if (x & 0x80000000)
            answer = ((long long)answer*(long long)y)%MODULUS;
        x <<= 1;
    }
    return answer;
}

/* The hashfunc calculates a hash based on the current board position.
 * The hash is a pseudo-random number which must be the same for any
 * two identical board positions.  In the game of Patience we consider
 * two board positions similar if they contain the same columns of cards
 * in any order, so the hash function is invariant when columns are swapped.
 *
 * The value of the hash will be (((a^col1)^col2)^...)^colN mod(MODULUS) and
 * then further modified by the cell card if any; where a is the RNG generator
 * and col1...colN are hash values computed from the card columns.  This
 * value is equivalent to a^(col1*col2*...*colN) and therefore the order of
 * the columns is not significant.  Each column hash must not be zero, and
 * so will be '1' if the column is empty.
 */
unsigned long hashfunc() {
    unsigned long ans=GENERATOR;   /* cumulative hash value */
    unsigned long colsum;          /* mini-hash of each column */
    int col, row;
    for (col=0; col<CELL; col++) { /* consider each column (not the cell) */
        colsum=1+board[col][0];    /* seed the column hash with no.of cards */
        for (row=board[col][0]; row>0; row--) {
            int card=board[col][row];
            if (card&REDCARD) {card -= (REDCARD-5);}  /* convert red cards to small integers 11-15 */
            colsum = colsum*75+card;  /* mix each card into the column hash */
        }
        ans = powmod(ans, 1+colsum);  /* mix each column hash into the main hash */
    }
    if (board[CELL][0]) {  /* mix the cell card into the final hash value */
        ans = (ans*75+board[CELL][1]) % MODULUS;
    }
    return ans;
}

/* The clearhash function (a) allocates the hash if it was not already, and
 * (b) frees all hash entries if the hash table had already been used.
 * Each hash entry is cleared to NULL to signify that it is unused.
 */
void clearhash() {
   int i;
   Hash *ptr,*nxt;
   hashsize=0;
   if (!hash) {
      hash=(Hash**)myalloc(HASHSIZE*sizeof *hash);
      for (i=0; i<HASHSIZE; i++) hash[i]=0;
      return;
   }
   for (i=0; i<HASHSIZE; i++) if (hash[i]) {
      ptr=hash[i]->next;
      free(hash[i]);
      hash[i]=0;
      while(ptr) {
         nxt=ptr->next;
         free(ptr);
         ptr=nxt;
      }
   }
}

/* The boardcmp function compares a stored board in the hash table with the
 * current board position, returning 0 if it matches and 1 if it doesn't match.
 * The board is considered to match if it contains the same columns in any
 * order.  The format of the board stored in the hash table is similar to
 * the current board array except that unused bytes are omitted.  So the
 * first byte is the number N0 of cards in the first column, and the next
 * N0 bytes are the cards; this is immediately followed by N1, the number
 * of cards in the second column, and so on.
*/
int boardcmp(const unsigned char *b) {
    unsigned char used[CELL];  /* we will mark off each column in the current
                                  board as it is matched with the stored one. */
    int i,j,k, ok;
    memset(used,0,CELL);       /* all columns start unused */
    for (i=0; i<CELL; i++) {   /* for each column in the stored board */
        for (j=0; j<CELL; j++) {   /* search for an identical column in the current board */
            ok=0;
            if (used[j]) continue; /* current board column must be unused */
            if (board[j][0] != b[0]) continue;  /* count of cards must match */
            ok=1;
            for (k=1; k<=board[j][0]; k++)      /* each individual card must match */
                if (board[j][k] != b[k]) {ok=0; break;}
            if (ok) break;
        }
        if (ok) {                  /* if an identical column was found */
            b+=b[0]+1;             /* consider next column in the hashed board */
            used[j]=1;             /* and mark the matched column used in the current board */
        }
        else return 1;             /* else column not found so the match failed */
    }
    /* when control reaches here, all columns matched, so we just have to
       compare the free cell */
    if (b[0]!=board[CELL][0]) return 1;
    if (b[0] && b[1]!=board[CELL][1]) return 1;
    /* and if that matched, the answer is success */
    return 0;
}

/* The following function is not used, but would have compared the board
 * literally rather than allowing the order of columns to differ.
 */
int boardcmp2(const unsigned char *b) {
    int i,j;
    for (i=0; i<=CELL; i++) {
        if (board[i][0] != b++[0]) return 1;
        for (j=1; j<=board[i][0]; j++)
            if (board[i][j] != b++[0]) return 1;
    }
    return 0;
}

/* some statistics for the hash table */
static unsigned long collisions=0;
static unsigned long hits=0;

/* The hashsearch function searches the cache for the current board
 * position given h, the hash of the current board position.
 * If found: returns 1 and sets *slot to point to the cache entry
 * otherwise: returns 0 and sets *slot to the address of the null
 * pointer that must be filled in when the hash entry is created.
 *
 * Both the hash and the board position must match (to eliminate false
 * matches due to hash collisions).  The hash h typically contains more
 * bits than the hash array index, so comparing it first eliminates some
 * collisions that were due to the hash table indexing fewer bits.
 */
int hashsearch(unsigned long h, Hash ***slot) {
    Hash **ptr=&hash[h&HASHMASK];  /* ptr is the address of a pointer to the entry */
    int c=0;                       /* c records whether there was an entry that didn't match */
    while (*ptr) {
        if (ptr[0]->hash == h && 0==boardcmp(ptr[0]->b)) {
            *slot=ptr;
            return 1;
        }
        c=1;
        ptr=&(ptr[0]->next);       /* ptr is the address of the next pointer */
    }
    /* if we reach here, a matching entry wasn't found. */
    *slot=ptr;
    if (c) collisions++;
    return 0;
}

static int clr=0;                  /* number of times the hash has been cleaned */

/* The hashput function adds a value to the cache based on the current 
 * board position.
 * h: the hash of the current board position
 * slot: the result from hashsearch when a search was unsuccessful
 * max: the maximum number of moves under consideration
 * result: the answer to be cached.
 * The cache entry will also contain a copy of the current board position.
 */
void hashput(unsigned long h, Hash **slot, int max, int result) {
    int i,l;
    Hash *entry;
    unsigned char *bptr;
    l=0;
    /* first calculate how many bytes are needed for the board, and allocate
       memory for the hash entry. */
    for (i=0; i<=CELL; i++) l+=board[i][0]+1;
    entry=myalloc(sizeof (Hash) + l);

    /* point the slot to the entry and fill in the values */
    slot[0]=entry;
    entry->next=0;
    entry->hash=h;
    entry->max=max;
    entry->result=result;
    /* add a copy of the board */
    bptr=entry->b;
    for (i=0; i<=CELL; i++) {
        bptr[0]=board[i][0];
        for (l=1; l<=board[i][0]; l++) bptr[l]=board[i][l];
        bptr+=board[i][0]+1;
    }
    /* count the entry and see if the table is considered too full */
    if (++hashsize == HASHFULL) {
        /* As operating on a hash table that has too many occupied slots can be
           inefficient, we will clear out some of the entries.
           This will free one in every four hash slots, and if it happens 
           several times will rotate the location of the first cleared entry
           among the first four entries in the hash, thus providing a sort of
           poor-man's least-recently-created algorithm.
        */
        Hash *ptr, *nxt;
        unsigned long c=0;
        fputs("Hash is full\n",stderr);
        for (i=clr++&3; i<HASHSIZE; i+=4) if (hash[i]) {
            ptr=hash[i]->next;
            free(hash[i]);
            hash[i]=0;
            c++;
            while(ptr) {
                nxt=ptr->next;
                free(ptr);
                ptr=nxt;
                c++;
            }
        }
        hashsize-=c;
        fprintf(stderr, "Cleared %lu entries (now %d%%)\n",c,(int)(hashsize*100/HASHFULL));
    }
}

/* The check function returns 1 if the board is in a solved position,
 * and 0 otherwise.
 */
int check() {
    int col;
    int first;
    int len;
    int i;
    for (col=0; col<CELL; col++) {
        len=board[col][0];  /* number of cards in each column - must be either */
        if (!len) continue; /* zero or 4 if the cards are suits or 5 otherwise */
        first=board[col][1];
        if (len != 4+(first>5)) return 0;
        if (first<6) {      /* if they are suits, all cards must be identical  */
            for (i=2; --len; i++) if(board[col][i]!=first) return 0;
        } else {            /* otherwise must decrease and alternate in colour */
            for (i=2; --len; i++) {
                first = (first-1)^REDCARD;
                if (board[col][i]!=first) return 0;
            }
        }
    }
    /* If nothing failed to match, the board is solved. */
    return 1;
}

/* The readboard function reads the input string representing a board
 * position and stores it in the current board.
 */
void readboard(const char *p) {
    static const char hexchars[]="0123456789abcdefABCDEF";
    int col,row;
    int num;
    int radix;
    char *h;
    /* first character must always be '[' */
    if (p++[0]!='[') {fputs("missing '['\n", stderr); exit(1);}
    for (col=0; col<=CELL; col++) {
        /* first character of each column must also be '[' */
        if (p++[0]!='[') {fputs("missing '['\n", stderr); exit(1);}
        row=1;
        /* we now expect zero or more numbers terminated by ']' */
        while(p[0]!=']'){
            num=0;
            /* number may start '0x' in which case it is in hex */
            if (p[0]=='0' && p[1]=='x') {
                p+=2;
                radix=16;
            }
            else radix=10; /* otherwise it's decimal */
            while ((h=strchr(hexchars,p[0]))) { /* read every valid digit */
                int digit=h-hexchars;
                if (digit>15) digit-=6;  /* capital letters same value as lowercase */
                if (digit>=radix) break; /* don't accept hex digit in a decimal number */
                num = num*radix + digit;
                p++;
            }
            board[col][row++]=num;       /* store the card */
            if (p[0]==',') {             /* skip comma followed by optional space */
                while(*++p==' ');
            } else {
                break;                   /* if no comma then it's end of list */
            }
        }
        /* end of the column list should have a ']' */
        if (p++[0]!=']') {fputs("missing ']'\n", stderr); exit(1);}
        board[col][0]=row-1;             /* store the count of cards in each row */
        if (p[0]==',') {                 /* skip comma followed by optional space */
            while(*++p==' ');
        } else {
            break;                       /* if no comma then it's end of list */
        }
    }
    /* the very end of the list should also have ']' but it's non-fatal */
    if (p++[0]!=']') {fputs("warning: missing ']'\n", stderr);}
}

/* The makemove function moves a card or sequence of cards from (x1,y1)
 * to (x2,y2).  The move is expected to be legal.  The row coordinates
 * are 1-based.
 */
void makemove(int x1, int y1, int x2, int y2) {
    int n;
    /* sanity check: y1 card must exist, and y2 must be exactly the
       first empty slot in the x2 column */
    if (!(y1 <= board[x1][0] && y2 == 1+board[x2][0])) {
        fputs("makemove: illegal move\n", stderr);
        exit(1);
    }
    n = board[x1][0];   /* save the current column count for x1 */
    board[x1][0]=y1-1;  /* delete the card(s) from column x1 */
    while (y1<=n) board[x2][y2++]=board[x1][y1++];  /* add the cards to column x2 */
    board[x2][0]=y2-1;  /* and set the new column count for x2 */
}

/* The solution will be built from end to start in the solution[] array which
   simply holds the x1,y1,x2,y2 of each move.  When the board is solved, the
   winning move will be placed at solution[d] where d is the recursion depth,
   and each "yes" return from the solve function will store another move in
   the previous position.  The soldepth variable holds the number of moves
   currently stored in the solution array. */
static int solution[MAXDEPTH][4];
static int soldepth=0;

/* For timing purposes we will store the time at which the solve started. */
static time_t starttime;

/* The matrix regulates whether moves are allowed in our restricted ruleset.
   matrix[i][j] is 1 if a move from col i to col j is allowed, 0 if not.
   We will start with every move allowed (except for moving a column to
   itself).  Moves will be disallowed and re-allowed as the game changes. 
   However, the matrix is only consulted if the cache is not in use, because
   entries in the cache should not depend on prior game state (and the cache
   will cut out all the repeated positions that the matrix is meant to
   prevent). */
static unsigned char matrix[CELL+1][CELL+1];
void initmatrix() {
    int i;
    memset(matrix, 1, sizeof matrix);
    for (i=0; i<=CELL; i++) matrix[i][i]=0;
}

/* For verbose printouts, if the most recent result was obtained via the hash
   then 'savedhash' will contain the hash value; otherwise savedhash=0. */
static unsigned long savedhash;

static unsigned int iter = 0; /* measure the number of times subsolve is called */

/* The subsolve function does the actual work of trying to solve the game from
 * the current position.
 * usecell: whether the free cell is allowed to be used
 * depth: the current recursion depth
 * maxmoves: maximum number of (remaining) moves permitted in any solution
 * return value: -1 if no solution found, or
                 the number of moves in the best solution
 */
int subsolve(int usecell, int depth, int maxmoves) {
    int x1, y1, x2, y2, n1;
    unsigned char card;
    unsigned long h;
    int r;
    int found=0;          /* whether a winning move has been found during this call */
    static int gfound=0;  /* whether a winning move has been found ever */
    Hash **hptr=0;        /* if cache is in use this will point to a slot */

    /* if we return a hashed result, 'savedhash' will be filled in; but 
       otherwise it should be zero. */
    savedhash=0;

    /* If the game is already solved, return success */
    if (check()) {
        soldepth=0;       /* there are no remaining moves in the solution */
        return 0;
    }

    /* If we are using the cache, return failure if the cache says that
       the game definitely can't be solved in the permitted number of moves */
    if (depth>cmindepth && depth<=cachedepth) {
        h = hashfunc();
        if (hashsearch(h,&hptr)) {
            r = hptr[0]->result;
            /* Board position was found in the cache; but the cached entry is
               only valid if it allowed at least 'maxmoves' moves.  Return
               failure if the cached value says the game can't be solved or
               takes more moves than permitted.  Otherwise, we will recompute
               the solution. */
            if (hptr[0]->max >= maxmoves) {
                if (r<0 || r > maxmoves) {hits++;savedhash=h;return -1;}
            }
            else {
                /* The cache entry didn't answer the question, so reinitialise */
                hptr[0]->max = maxmoves;
                hptr[0]->result = -1;
            }
        }
        /* If the current position was not cached, add a cache entry.  The
           new entry will indicate failure (until it's eventually filled in
           at the end of this search); this will cut short any future searches
           if we somehow arrive at the same board position later on. */
        else hashput(h, hptr, maxmoves, -1);
    }

    /* Before starting the search for a winning move, check whether we have
       timed out.  Avoid calling time() too much by only doing this every
       10000 iterations.  If we exceeded the maximum time, return failure;
       if we exceeded the optimisation time and a solution has been found,
       also return failure (and the caller will use the most recent solution
       it did find).  Because the iteration count hasn't been incremented yet,
       when a timeout occurs all future calls here will also time out. */
    if (iter%10000 == 0) {
        time_t secs=time(NULL)-starttime;
        if (secs >= MAXTIME) {MAXTIME=0; return -1;}
        if (secs >= OPTTIME && gfound) {OPTTIME=0; return -1;}
    }
    iter++; /* Now increment the interation count. */

    /* Begin the search for a winning move, starting in the leftmost column */
    for (x1=0; x1<=CELL; x1++) {
        if (x1==CELL && !usecell) break; /* don't use the cell if not permitted */
        y1=board[x1][0];
        if (y1==0) continue;   /* can't use this column if there are no cards */
        /* if column x1 contains a sequence of cards, we will move the whole
           sequence, so look for a sequence.  n1 is how many cards will be moved. */
        n1=1;              
        card = board[x1][y1];
        while (y1>1) {
            if (card<6 ? card!=board[x1][y1-1] : card != ((board[x1][y1-1]-1)^REDCARD))
                break;
            y1--;
            n1++;
            card=board[x1][y1];
        }
        /* Start searching for a column to move to, starting with the leftmost */
        for (x2=0; x2<=CELL; x2++) {
            /* If the cache is in use, the matrix will not be consulted; but
               don't try moving a column to itself.  Otherwise, use the matrix
               to find out whether this move should be tested (the matrix
               automatically prohibits moving a column to itself). */
            if (hptr) {if (x1==x2) continue;}
            else if (!matrix[x1][x2]) continue;
            matrix[x1][x2]=0;   /* disallow testing this same move in future */

            if (x2==CELL) {
                /* If moving to the cell: don't do it if the cell is disallowed,
                   or the cell is full, or we are trying to move a sequence.
                   Also don't bother if a winning move has already been found. */
                if (!usecell || n1>1 || board[CELL][0]>0 || found)
                    break;
            }
            /* Now check that x1 to x2 is a legal move */
            y2=board[x2][0]+1;
            if (y2>1) {
                if (card<6 ? card!=board[x2][y2-1] : card!=((board[x2][y2-1]-1)^REDCARD))
                    continue;
            }
            else if (y1==1) continue; /* don't move a whole column to another empty column */

            /* We are searching for a solution better than 'maxmoves', so only
               go ahead if there are moves left.  In that case we will try this
               move and call the solver recursively to see if there is a
               solution from there. */
            if (maxmoves>n1) {
                int moves;
                unsigned char savematrix[CELL+1][CELL+1];
                int i;
                /* We are about to mangle the matrix, so save a copy */
                memcpy(savematrix,matrix,sizeof matrix);
                /* When we move x1 to x2 then all further moves from x2 are
                   banned except for x2 to x1.  All further moves from and to x1
                   are allowed because the column status changed.  And all
                   moves to x2 are also allowed.
                   When we move x1 to cell and x1 is now empty, cell is banned
                   from moving to any empty column.
                   When we move cell to x2 [and x2 was empty], x2 is banned
                   from moving back to cell.  [If x2 was not empty then all
                   further moves from x2 will involve a sequence and thus
                   cannot be moved back to cell anyway.] */
                if (x2==CELL) {
                    for (i=0; i<CELL; i++) if (i!=x1) matrix[i][x1]=matrix[x1][i]=1;
                    if (y1 == 1) {
                        matrix[CELL][x1]=0;
                        for (i=0; i<CELL; i++) if (board[i][0]==0) matrix[CELL][i]=0;
                    }
                }
                else if (x1==CELL) {
                    for (i=0; i<CELL; i++) if (i!=x2) matrix[i][x2]=1;
                    matrix[x2][CELL]=0;
                }
                else {
                    for (i=0; i<=CELL; i++) matrix[x2][i]=0;
                    for (i=0; i<=CELL; i++) if (i!=x1) matrix[i][x1]=matrix[x1][i]=1;
                    for (i=0; i<=CELL; i++) if (i!=x2) matrix[i][x2]=1;
                }

                /* Now make the move that we want to test */
                makemove(x1, y1, x2, y2);
                if (verbose>1 && depth<=vdepth) {
                    int j=depth;
                    while(j--) putc(' ', stderr);
                    fprintf(stderr, "try (%d,%d) to (%d,%d) [%lx]\n", x1, y1, x2, y2, hashfunc());
                }
                /* and call the solver recursively to see if this is a winning position */
                moves = subsolve(usecell, depth+1, maxmoves-n1);
                if (verbose>1 && depth<=vdepth) {
                    int j=depth;
                    while(j--) putc(' ', stderr);
                    fprintf(stderr, "<- (%d,%d) to (%d,%d) = %d", x1, y1, x2, y2, moves);
                    if (savedhash) fprintf(stderr, " (cached %06lx)", savedhash);
                    fputc('\n',stderr);
                }
                /* Undo the move and retrieve the original matrix */
                makemove(x2, y2, x1, y1);
                memcpy(matrix, savematrix, sizeof matrix);

                /* now test the result from the solver */
                if (moves>=0 && moves+n1 < maxmoves) {
                    /* the solver found a winning sequence that was better
                       than the current best, so update the max moves we are
                       looking for and record the current move in the solution */
                    maxmoves=moves+n1;
                    soldepth++;
                    solution[depth][0]=x1;
                    solution[depth][1]=y1;
                    solution[depth][2]=x2;
                    solution[depth][3]=y2;
                    found=gfound=1;
                }
                /* To cut down the search space, we will not try moving a card
                   to the cell if the card has any other legal move */
                matrix[x1][CELL]=0;
            }
        }
    }
    /* Search has finished: time to return the result */
    savedhash=0;  /* signal that the result wasn't cached */
    if (found) {
        /* a winning move was found: update the cache and return the result */
        if (hptr && hptr[0]->result==-1) hptr[0]->result=maxmoves;
        return maxmoves;
    }
    /* otherwise return failure */
    return -1;
}

/* The solve function simply has to call subsolve to find a solution.
 * It does this twice: once without using the free cell, and (if that failed)
 * once with the cell.  This function also takes care of initialising the
 * hash and setting up the timing.
 */
int solve() {
    int moves;
    initmatrix();
    clearhash();
    starttime = time(NULL);
    moves = subsolve(0, 0, MAXMOVES);
    if (moves>0) {
        if (OPTTIME==0 && verbose) fputs("(optimisation timed out - using best so far)\n",stderr);
        if (verbose) fprintf(stderr, "Without using the free cell: ");
    }
    else if (MAXTIME>0) {
        int secs=time(NULL)-starttime;
        if (verbose) fprintf(stderr, "Enabling free cell after %d secs\n", secs);
        initmatrix();
        clearhash();
        starttime = time(NULL);
        moves = subsolve(1, 0, MAXMOVES);
        if (OPTTIME==0 && verbose) fputs("(optimisation timed out - using best so far)\n",stderr);
    }
    if (MAXTIME==0 && verbose)
        fputs("Timed out\n", stderr);
    return moves;
}

/* Finally, the main program which parses the arguments, calls the solver,
 * and prints out the result.
 */
int main(int argc, char **argv) {
    int m;
    int secs;
    int i;
    for (i=1; argc>i && argv[i][0]=='-'; i++) { /* collect flag arguments */
        switch(argv[i][1]) {
            case 'v': verbose=1;
                if (argv[i][2]=='v') {
                    int d = atoi(argv[i]+3);
                    verbose++;
                    if (d>0) vdepth=d;
                }
                break;
            case 'n': cmindepth=100;
                break;
            default: fprintf(stderr,"Incorrect flag %s\n", argv[i]);
        }
    }
    if (argc!=i+1) {fputs("Incorrect parameters\n",stderr); exit(1);}

    readboard(argv[i]); /* collect the board */
    m=solve();          /* solve the board */
    secs=time(NULL)-starttime;
    if (m<0) {
        if (verbose) fprintf(stderr, "No solution found (%d secs)\n",secs);
        puts("None");
    }
    else {
        int i;
        if (verbose) fprintf(stderr, "Solution found in %d secs - %d moves (%d cards)\n", secs, soldepth, m);
        if (verbose) fprintf(stderr, "Cache size %lu - hits %lu collisions %lu\n", hashsize, hits, collisions);
        /* Now print out the solution.  Our rows are 1-based, but the printout
           will be 0-based so subtract 1 from each row value. */
        putchar('[');
        for (i=0; i<soldepth; i++) {
            if (i) printf(", ");
            printf("[(%d,%d), (%d,%d)]", solution[i][0], solution[i][1]-1, solution[i][2], solution[i][3]-1);
        }
        putchar(']');
        putchar('\n');
    }
    return 0;
}
